/*
 *  YASS 1.09
 *  Copyright (C) 2004,2005
 *  YASS Team
 *  Laurent Noe, Gregory Kucherov, Mikhail Roytberg, Christophe Valmir, Steven Corroy
 *  yass@loria.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software contains code derived from the GNU libavl library.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>


#include "kword.h"
#include "tuple.h"
#include "prdyn.h"
#include "proba.h"
#include "align.h"
#include "assemble.h"
#include "display.h"
#include "global_var.h"
#include "hash.h"
 
/*
 * Hach table division
 */
#define DIV_HACH 135
#define H_DIST(distance) ((distance)/DIV_HACH)

/*
 * Modulo access to the last tuple
 */

#ifdef MEMOPT
  /* query table index (little more time consuming but memory economic ) */
  int memopt_modmask = 0;
  #define MEMOPT_SETMOD   {int i=0;memopt_modmask = 0;for(i=0; ((keysize_query + gp_rho_stat + gp_delta_stat + 1) >= memopt_modmask);i++)  memopt_modmask |=  (1 << i);}
  #define MEMOPT_MOD      &(memopt_modmask)
  #define MEMOPT_TABSIZE   (memopt_modmask+1)
#else
  /* full table index (memory consuming) */ 
  #define MEMOPT_SETMOD
  #define MEMOPT_MOD       
  #define MEMOPT_TABSIZE   (keysize_query + keysize_text)
#endif

/*
 * Shift tables used in assemble algorithms
 */

int * deltashift = NULL;


int initialise_deltashift()
{
  int i,s=1;
  deltashift = (int * )malloc((3+2*gp_delta_stat)*sizeof(int));
  ASSERT(deltashift,"initialise_deltashift");
  for ( i=0 ; i<3+2*gp_delta_stat ; i++) {
    deltashift[i] = s * (i+1) / 2;
    s *= -1;
  }
  return 0;
}

/*
   definition des fonctions de la table hash ou tableaux
 */
#ifdef LOW_MEMORY
#define INIT_TAB(table,type,size,set)     table=initHash(gp_hash_size,gp_hash_element)
#define GET_TAB_MIN(table,index,minValue) getHashValueMin(&table,index,minValue)
#define GET_TAB(table,index)              getHashValue(&table,index)
#define EXIST_TAB(table,index)            existHashDiag(&table,index)
#define PUT_TAB(table,index,value)        putHashOneDiagValue (&table,index,(int)value)
#define DEL_TAB(table,index)              delHash (&table,index)
#define FREE_TAB(table)                   freeHash(&table)
#define RESET_TAB(table,size,set)         resetHash(&table);
#else
#define INIT_TAB(table,type,size,set) table = (type) malloc(size );\
                                 if(table==NULL){                 \
                                   fprintf(stderr,"Assemble: can not memory, try './configure --with-low-memory'\n"); \
                                   exit(1);                       \
                                 }                                \
                                 memset(table,set,(unsigned)(size));

#define GET_TAB_MIN(table,index,minValue) table[index MEMOPT_MOD]
#define GET_TAB(table,index)            table[index MEMOPT_MOD]
#define EXIST_TAB(table,index)          table[index MEMOPT_MOD]
#define PUT_TAB(table,index,value)      table[index MEMOPT_MOD]=value
#define DEL_TAB(table,index)            table[index MEMOPT_MOD]=NULL
#define FREE_TAB(table)                 free(table)
#define RESET_TAB(table,size,set)       memset(table,size,(unsigned)set);
#endif

int Assemble_Single(/*in*/ char * data, /*in*/ int datasize,
		    Feature_assemble  *feature)
{
   
  /* main variables 
   * for the linked list computation
   */
  int * keylist[MAX_SEED];
  int * first[MAX_SEED];
  int keysize_query = datasize - gp_seeds_span_min + 1;
  register int index=0;
  register int index_fin=0;
  
  /* for the linking algorithm */
  int i_previous;
  int j;
  int distance,observed_distance,delta_distance;
   
  /* auxiliary variables */
  int code;
#ifdef LOW_MEMORY
  /* use hash table*/
  Table_hash  last_tuple_pos_with_dist;
  Table_hash  last_tuple_ptr_with_dist;
#else
  int  *   last_tuple_pos_with_dist = NULL;
  tuple ** last_tuple_ptr_with_dist = NULL;
#endif

  /* For multi seed */
  int seed;

  /* Cache */
#ifndef NO_CACHE
  int j_hack; 
  register int borne_min_hash=gp_rho_stat/2-1;
  register int borne_max_hash=DIV_HACH-gp_rho_stat/2+1;
  int *last_tuple_pos_with_dist_hach;
#endif


  /* sequence covering variables */ 
  tuple     * t, * t_;

  /* [0] left correction factor set to zero */
  feature->left_correction =  0;

  /* to keep lists of tuple found */
  feature->tuple_lists = CreateTupleList();
  feature->last_tl     = feature->tuple_lists;

  /* [1] key linked list creation */
  for( seed= 0 ; seed < gp_nb_seeds ; seed ++ ) 
    CreateKeyList(data,datasize,&first[seed],&keylist[seed],seed);



#ifdef  DEBUG 
  fprintf (stderr,"span=%d  gp_rho_stat=%d gp_distdiag=%d gp_rho_stat=%d gp_delta_stat=%d gp_indels=%d\n",gp_seeds_span[0],gp_delta_stat,gp_distdiag,gp_rho_stat,gp_delta_stat,gp_indels);
  fprintf(stderr," motif:%s\n",gp_motifs[0]);
  for(index=0;index<datasize - gp_seeds_span[seed] + 1;index++) {
    fprintf(stderr,"first   %d %d\n",index,first[0][index]);
    fprintf(stderr,"keylist %d %d\n",index,keylist[0][index]);
  }

#endif

  /* [2] allocations */
  INIT_TAB(last_tuple_pos_with_dist,int *,keysize_query * sizeof(int),0xff);
  INIT_TAB(last_tuple_ptr_with_dist,tuple **,keysize_query * sizeof(tuple *),0x00);

  

#ifndef NO_CACHE
  last_tuple_pos_with_dist_hach = (int *) malloc((unsigned)(((keysize_query+keysize_query)/DIV_HACH+1) * sizeof(int)));
  ASSERT(last_tuple_pos_with_dist_hach,"Assemble");
  memset(last_tuple_pos_with_dist_hach,0x0,(unsigned)(((keysize_query+keysize_query)/DIV_HACH+1) * sizeof(int)));
#endif

  
  
  
  /* [3] algorithm to create linked tuples */

#ifdef STATS
  last_clock = clock();
#endif
  /* for each kwork of the Text */
  for(feature->i_current = 0 ; feature->i_current < keysize_query ; feature->i_current++ ) {
    for( seed= 0 ; seed < gp_nb_seeds ; seed ++ ) 
      if ( feature->i_current<datasize - gp_seeds_span[seed] + 1){
 

      code      = Key(data,feature->i_current,seed); /* we take the code of the kmot 'w' located at i_current position     */
      index     = first[seed][code];  /* search the first occurrence of 'w' on the sequence */
      index_fin = first[seed][code+1];
      i_previous= keylist[seed][index];


      /* consider all the occurrences of 'w' before i_current */
      while(i_previous < feature->i_current - gp_distdiag && index<index_fin ){
	distance = feature->i_current - i_previous;
	/* consider all the interesting distances (diagonals) by adding indel bias  0 +1 -1 +2... until we reach a new tuple or the upper indel bound */
	observed_distance = distance;
	delta_distance = 0;

#ifndef NO_CACHE 
	/* Jump if no match on the breigthour */
	j_hack=last_tuple_pos_with_dist_hach[H_DIST(observed_distance )]; 
	code=((j_hack< feature->i_current - gp_rho_stat)||(j_hack > feature->i_current));
	if (observed_distance%DIV_HACH>borne_max_hash) {
	  j_hack=last_tuple_pos_with_dist_hach[H_DIST(observed_distance)+1]; 
	  code=code&&((j_hack< feature->i_current - gp_rho_stat)||(j_hack > feature->i_current)); 
	}
	if (observed_distance%DIV_HACH<borne_min_hash) {
	  j_hack=last_tuple_pos_with_dist_hach[H_DIST(observed_distance)-1]; 
	  code=code&&((j_hack< feature->i_current - gp_rho_stat)||(j_hack > feature->i_current));
	} 
        if (code==0)
#endif

	while( delta_distance <= 2 * gp_delta_stat + 1){
	  
	  /* search for interesting tuples */	  
	  j = GET_TAB_MIN(last_tuple_pos_with_dist,observed_distance,gv_i_current - gp_rho_stat);

	  /* if tuples found are respecting criteria (neighboors on diagonal and on sequences) */
	  if ( j > feature->i_current - gp_rho_stat                   &&
	       j <= feature->i_current                                &&
	       j - observed_distance <= feature->i_current - distance &&
	       j != -1 ){


	    /* hachDisplay (&last_tuple_ptr_with_dist); */
	    /*---------------------------------------------------------------------*/
	    
	    /* (1) possibly create one instance of tuple [j]
	     * if this one was not in any list.
	     */
	    if (EXIST_TAB(last_tuple_ptr_with_dist,observed_distance)<=0){
	      
#ifdef STATS
		feature->nb_chains_built+=1;
#endif
		
		/* create at tuple list element */
		feature->last_tl->next = CreateTupleList();
		feature->last_tl = feature->last_tl->next;
		
		/* create a tuple [j] and update lists */
		t  = CreateTuple(j,observed_distance,seed);
		t->bitweight=gp_automata[gp_seeds_bitweight[seed]];
		t->overlappedhits=gp_seeds_bitweight[seed];
		feature->last_tl->first_tuple = t;
		PUT_TAB(last_tuple_ptr_with_dist,observed_distance,t);
	    }
	    else
	      /* (2) we search for the last tuple [j] instanciation
	       */
	      t = (tuple *)GET_TAB(last_tuple_ptr_with_dist,observed_distance) ;
	      
	      /*
	       * (3) iterate tuples and add this new one to the most adapted one
	       */
	      
	      do{
		
		if (  distance == t->distance  && feature->i_current - t->occurrence <= gp_seeds_span[seed] ){

		  /* (3a) possible fusion with this tuple (no instanciation)
		   *
		   * [inter-tuple overlap]
		   */ 

		  t->size              +=  feature->i_current -  t->occurrence;
		  t->lastSeed           = seed;
		  if ((t->lastSeed==-1) ||(gp_seeds_span[t->lastSeed]<gp_seeds_span[seed]))
		    t->lastSeed           = seed;
		  t->bitweight         += gp_automata[(~(t->overlappedhits<<(feature->i_current -  t->occurrence)))&gp_seeds_bitweight[t->lastSeed]];
		  t->overlappedhits    = t->overlappedhits<<(feature->i_current -  t->occurrence) | gp_seeds_bitweight[t->lastSeed];
		  t->overlappedhits    &= AUTOMATAMASK; 
		  t->occurrence         = feature->i_current;
		  goto no_maj_last_tuple_ptr_with_dist;
		  
		}
		
		/* list searching for the most fitted */
		if(!(t->next))
		  break;
		t = t->next;
		
	      }while(1);
	  
	      /*
	       * (3d) any fusion has not been possible
	       *
	       * we instanciate the tuple and add this one at the end of the chain.
	       *
	       */
	      t_ = CreateTuple(feature->i_current, distance, seed);
	      t->next = t_;
	      PUT_TAB    (last_tuple_ptr_with_dist,distance,t_);
	      
	      /*
	       * (4) dont reset last_tuple_ptr_with_dist[distance] to NULL
	       * because it has been modified in (3)
	       */
	      goto no_maj_last_tuple_ptr_with_dist;	  
	    }

	  /*---------------------------------------------------------------------*/
	  /* diagonal walk d,d+1d-1,... */
	  delta_distance++;
	  observed_distance = distance + deltashift[delta_distance];
	  observed_distance = MAX(MIN(observed_distance,keysize_query-1),1);
	  /* border effects avoided*/
	  
	}/* while : diagonal walk */
	
	
	/* reset last tuple instancied to NULL (because not instance is created)
	 */
	DEL_TAB    (last_tuple_ptr_with_dist,distance);	
	/* last_tuple_ptr_with_dist[distance] = NULL; */
	
      no_maj_last_tuple_ptr_with_dist:
	
	/*
	 * update last tuple position  with diagonal = 'distance'
	 */
#ifdef STATS
	feature->nb_seeds++;
#endif
	PUT_TAB(last_tuple_pos_with_dist,distance,feature->i_current);

#ifndef NO_CACHE
	last_tuple_pos_with_dist_hach[H_DIST(distance)]=feature->i_current;
#endif
	/* iterate 'w' code list before feature->i_current position */
	index ++;
	i_previous = keylist[seed][index]; 	

        }  

      }/*end of seed */ 
	

      if (!(feature->i_current % ALIGN_EVERY_NB_ITER)){
#ifdef TRACE 
	Display_Progress((unsigned)feature->i_current,(unsigned)(keysize_query - 1),feature);
#endif 

#ifdef STATS
	feature->clock_chain += clock() - last_clock;
	last_clock=  clock();
#endif		
#ifdef DEBUGASSEMBLE
	DisplayListTupleList(feature->tuple_lists);
#endif
	AlignAndFree( 
		     data, datasize,
		     data, datasize,
		     0,0,
		     feature
		     );

#ifdef STATS
	feature->clock_align += clock() - last_clock;
	last_clock=  clock();
#endif
      }
      
  }/*end of [3] */
#ifdef TRACE
  Display_Progress((unsigned)(unsigned)feature->i_current,keysize_query - 1,feature);
#endif  
 
#ifdef STATS
  feature->clock_chain += clock() - last_clock;
  last_clock=  clock();
#endif	

  /* on aligne les tuples restants */
#ifdef DEBUGASSEMBLE
  DisplayListTupleList(tuple_lists,feature);
#endif
  AlignAndFree(
	       data, datasize,
	       data, datasize,
	       1,0,
	       feature
	       );
#ifdef STATS	
   feature->clock_align += clock() - last_clock;
   last_clock =  clock();
#endif

   FREE_TAB(last_tuple_pos_with_dist);
   FREE_TAB(last_tuple_ptr_with_dist);
   for( seed= 0 ; seed < gp_nb_seeds ; seed ++ ) {
     free(keylist[seed]);
     free(first[seed]);
   }
   return  0;
}


int Assemble_SingleRev(/*in*/   char * datarev, /*in*/  char * data, /*in*/ int datasize,
				Feature_assemble  *feature) {
  


 /* main variables */
  /* for the linked list computation */
  int *keylist[MAX_SEED];
  int *first  [MAX_SEED];
  int keysize_query = datasize - gp_seeds_span_min + 1;

  
#ifndef MEMOPT	
  int keysize_text  = keysize_query ; 
#endif 
  /* for the linking algorithm */
  int index,index_fin;
  int i_previous;
  int j;
  int distance,observed_distance,delta_distance;
   
  /* auxiliary variables */
  int code;
#ifdef LOW_MEMORY
  /* use hash table*/
  Table_hash  last_tuple_pos_with_dist;
  Table_hash  last_tuple_ptr_with_dist;
#else
  int  *   last_tuple_pos_with_dist = NULL;
  tuple ** last_tuple_ptr_with_dist = NULL;
#endif

  /* For multi seed */
  int seed;

  /* Cache */
#ifndef NO_CACHE
  int j_hack; 
  register int borne_min_hash=gp_rho_stat/2-1;
  register int borne_max_hash=DIV_HACH-gp_rho_stat/2+1;
  int *last_tuple_pos_with_dist_hach;
#endif

  /* sequence covering variables */ 
  tuple     * t, * t_;

  /* to keep lists of tuple found */
  feature->tuple_lists = CreateTupleList();
  feature->last_tl     = feature->tuple_lists;


  /* left correction factor set to keysize_query */
  feature->left_correction =  keysize_query;
  
  MEMOPT_SETMOD;  
 
  /* [1] key linked list creation */
  for (seed=0;seed<gp_nb_seeds;seed++)
    CreateKeyList(datarev,datasize,&first[seed],&keylist[seed],seed);

  /* [2] allocations */
  INIT_TAB(last_tuple_pos_with_dist,int *,MEMOPT_TABSIZE * sizeof(int),0xff);
  INIT_TAB(last_tuple_ptr_with_dist,tuple **,MEMOPT_TABSIZE * sizeof(int),0x00);

#ifndef NO_CACHE
  last_tuple_pos_with_dist_hach = (int *) malloc((unsigned)(((keysize_query+keysize_query)/DIV_HACH+1) * sizeof(int)));
  ASSERT(last_tuple_pos_with_dist_hach,"Assemble");
  memset(last_tuple_pos_with_dist_hach,0x0,(unsigned)(((keysize_query+keysize_query)/DIV_HACH+1) * sizeof(int)));
#endif
   
  /* [3] algorithm to create linked tuples */

#ifdef STATS
  last_clock = clock();
#endif
  
  /* for each kwork of the Text */
  for(feature->i_current = 0 ; feature->i_current < keysize_query ; feature->i_current ++ )
    for (seed=0;seed<gp_nb_seeds;seed++){
      
   
      code      = Key(data, feature->i_current,seed); /* we take the code of the kmot 'w' located at feature->i_current position     */
      index     = first[seed][code];  /* search the first occurrence of 'w' on the sequence */
      index_fin = first[seed][code+1];
      i_previous= keylist[seed][index];
      
      
      /* consider all the occurrences of 'w' before feature->i_current */
   
      
      while(i_previous < keysize_query - feature->i_current  && index<index_fin){  
	
	distance =   feature->i_current - i_previous + keysize_query ;      
	/* consider all the interesting distances (diagonals) by adding indel bias  0 +1 -1 +2... until we reach a new tuple or the upper indel bound */
	observed_distance = distance;
	delta_distance = 0;
	
#ifndef NO_CACHE 
	/* Jump if no match on the breigthour */
	j_hack=last_tuple_pos_with_dist_hach[H_DIST(observed_distance )]; 
	code=((j_hack< feature->i_current - gp_rho_stat)||(j_hack > feature->i_current));
	if (observed_distance%DIV_HACH>borne_max_hash) {
	  j_hack=last_tuple_pos_with_dist_hach[H_DIST(observed_distance)+1]; 
	  code=code&&((j_hack< feature->i_current - gp_rho_stat)||(j_hack > feature->i_current)); 
	}
	if (observed_distance%DIV_HACH<borne_min_hash) {
	  j_hack=last_tuple_pos_with_dist_hach[H_DIST(observed_distance)-1]; 
	  code=code&&((j_hack< feature->i_current - gp_rho_stat)||(j_hack > feature->i_current));
	} 
        if (code==0)
#endif

	while( delta_distance <= 2 * gp_delta_stat + 1){
	  
	  /* search for interesting tuples */
	  j = GET_TAB_MIN(last_tuple_pos_with_dist,observed_distance,gv_i_current - gp_rho_stat);
	  
	  /* if tuples found are respecting criteria (neighboors on diagonal and on sequences) */
	  if ( j > feature->i_current - gp_rho_stat                  &&                                         
	       j <= feature->i_current                                && 
	       j - observed_distance <= feature->i_current - distance && 
	       j != -1 ){ 
	  
	    
	    /*---------------------------------------------------------------------*/
	    
	    /* (1) possibly create one instance of tuple [j] 
	     * if this one was not in any list. 
	     */
	    
	    if (EXIST_TAB(last_tuple_ptr_with_dist,observed_distance)<=0){
            
#ifdef STATS
	      feature->nb_chains_built+=1;
#endif
	      
	      /* create at tuple list element */
	      feature->last_tl->next = CreateTupleList();
	      feature->last_tl = feature->last_tl->next;
	      
	      /* create a tuple [j] and update lists */
	      t  = CreateTuple(j,observed_distance,seed);
	      
	    
	      feature->last_tl->first_tuple = t;
	      PUT_TAB(last_tuple_ptr_with_dist,observed_distance,t);
	    }
	    else
	    /* (2) we search for the last tuple [j] instanciation 
	     */
	      t = (tuple *)GET_TAB(last_tuple_ptr_with_dist,observed_distance) ;

	    
	    /*
	     * (3) iterate tuples and add this new one to the most adapted one
	     */
	    
	    do{
	      
	      if (  distance == t->distance  && feature->i_current - t->occurrence <= gp_seeds_span[seed] ){
		
	      /* (3a) possible fusion with this tuple (no instanciation)
	       *	
	       * [inter-tuple overlap]	
	       */      
	       
		t->size              +=  feature->i_current -  t->occurrence;
		t->lastSeed           = seed;
		if ((t->lastSeed==-1) ||(gp_seeds_span[t->lastSeed]<gp_seeds_span[seed]))
		  t->lastSeed           = seed;
		t->bitweight         += gp_automata[(~(t->overlappedhits<<(feature->i_current -  t->occurrence)))&gp_seeds_bitweight[t->lastSeed]];
		t->overlappedhits    = t->overlappedhits<<(feature->i_current -  t->occurrence) | gp_seeds_bitweight[t->lastSeed];
		t->overlappedhits    &= AUTOMATAMASK; 
		t->occurrence         = feature->i_current;
		goto no_maj_last_tuple_ptr_with_dist;
		
	    }
	      
	    /*	parcours de la liste */
	    if(!(t->next))
	      break;
	    t = t->next;
	    
	  }while(1);
	  
	  /*
	   * (3d) any fusion has not been possible
	   *
	   * we instanciate the tuple and add this one at the end of the chain.
	   *
	   */
	  
	  t_ = CreateTuple(feature->i_current, distance,seed);
	  t->next = t_;
	  PUT_TAB    (last_tuple_ptr_with_dist,distance,t_);
	  
	  /*
	   * (4) dont reset last_tuple_ptr_with_dist[distance] to NULL 
	   * because it has been modified in (3)
	   */
	  goto no_maj_last_tuple_ptr_with_dist;
	  
	}
	
	/*---------------------------------------------------------------------*/
	

	/* diagonal walk d,d+1,d-1,... */
	delta_distance ++;
	observed_distance = distance + deltashift[delta_distance];
	observed_distance = MAX(MIN(observed_distance,keysize_query-1),1);
	/* border effects avoided*/
	  
      }/*  while : diagonal walk */
	

      /* reset last tuple instancied to NULL (because not instance is created) 
       */ 
	DEL_TAB    (last_tuple_ptr_with_dist,distance);	
	    
no_maj_last_tuple_ptr_with_dist:

      /* 
       * update last tuple position  with diagonal = 'distance'
       */
#ifdef STATS
      feature->nb_seeds++;
#endif
	PUT_TAB(last_tuple_pos_with_dist,distance,feature->i_current);
#ifndef NO_CACHE
	last_tuple_pos_with_dist_hach[H_DIST(distance)]=feature->i_current;
#endif
      /* iterate 'w' code list before feature->i_current position */
      index ++;

      i_previous = keylist[seed][index]; 
    }  
	
    if (!(feature->i_current % ALIGN_EVERY_NB_ITER)){
#ifdef TRACE
      Display_Progress((unsigned)feature->i_current,(unsigned)(keysize_query - 1),feature);
#endif
#ifdef STATS
      feature->clock_chain += clock() - last_clock;
      last_clock=  clock();
#endif		
#ifdef DEBUGASSEMBLE
      DisplayListTupleList(tuple_lists);
#endif
      AlignAndFree(
		   datarev, datasize,
		   data,    datasize,
		   0,0,
		   feature
		   );

#ifdef STATS

      feature->clock_align += clock() - last_clock;
      last_clock=  clock();
#endif
    }
    
  }/*end of [3] */
#ifdef TRACE
  Display_Progress((unsigned)feature->i_current,(unsigned)(keysize_query - 1),feature);
#endif  
 
#ifdef STATS
  feature->clock_chain += clock() - last_clock;
  last_clock=  clock();
#endif	

  /* on aligne les tuples restants */
#ifdef DEBUGASSEMBLE
  DisplayListTupleList(tuple_lists);
#endif
  AlignAndFree(
	       datarev,    datasize,
	       data, datasize,
	       1,0,
	       feature);
#ifdef STATS	
   feature->clock_align += clock() - last_clock;
   last_clock =  clock();
#endif

   FREE_TAB(last_tuple_pos_with_dist);
   FREE_TAB(last_tuple_ptr_with_dist);

   for(seed=0;seed<gp_nb_seeds;seed++) {
     free(keylist[seed]);
     free(first[seed]);
   }
   return  0;
}




int MultiAssemble_Double(char * data_query, int datasize_query,
			 char * data_text,  int datasize_text, 
			 int nbchunks_text, char ** chunkname_text,
			 int * chunksize_text,
			 int * chunkstrt_text,
			 Feature_assemble  *feature ){
    /* used for chunk selection */
  int i_chunk                    = 0;
  int i_base                     = 0;
  int max_chunk_datasize_text    = 0;
  int keysize_text     = 0;
  
  int     chunk_keysize_text     = 0;
  char *  chunk_data_text        = data_text;
  double last_clock;
  
  /*  main variables 
   *  for the linked list computation
   */
  
  int * keylist_query[MAX_SEED];
  int * first_query  [MAX_SEED];
  int   keysize_query = datasize_query - gp_seeds_span_min + 1;
  
  /* for the linking algorithm */
  int i_previous;
  int index,index_fin;
  int j;
  int distance,observed_distance,delta_distance;
  
  
  /* auxiliary variables */
  int      code;
#ifdef LOW_MEMORY
  /* use hash table*/
  Table_hash  last_tuple_pos_with_dist;
  Table_hash  last_tuple_ptr_with_dist;
#else
  int  *   last_tuple_pos_with_dist = NULL;
  tuple ** last_tuple_ptr_with_dist = NULL;
#endif 

  /* sequence covering variables */ 
  tuple     * t, * t_;

  /* Multi seed */
  int seed;

  /* Hach table */
#ifndef NO_CACHE 
  int j_hack; 
  register int borne_min_hash=gp_rho_stat/2-1;
  register int borne_max_hash=DIV_HACH-gp_rho_stat/2+1;
  int *last_tuple_pos_with_dist_hach;
#endif

  /* to keep lists of tuple found */
  feature->tuple_lists = CreateTupleList();
  feature->last_tl     = feature->tuple_lists;

  
  for(i_chunk = 0 ; i_chunk < nbchunks_text; i_chunk++)
    max_chunk_datasize_text = MAX(max_chunk_datasize_text, chunksize_text[i_chunk]);
  keysize_text = max_chunk_datasize_text - gp_seeds_span_min + 1; 


  /* [0] left correction factor set to keysize_query */
  feature->left_correction =  keysize_query;
  MEMOPT_SETMOD;

  /* [1] key linked list creation */
  for(seed=0;seed<gp_nb_seeds;seed++) {
    CreateKeyList(data_query,datasize_query,&first_query[seed],&keylist_query[seed],seed);
  }

  /* [2] allocations */
  INIT_TAB(last_tuple_pos_with_dist,int *,MEMOPT_TABSIZE * sizeof(int),0xff);
  INIT_TAB(last_tuple_ptr_with_dist,tuple **,MEMOPT_TABSIZE * sizeof(int),0x00);

#ifndef NO_CACHE 
  last_tuple_pos_with_dist_hach = (int *) malloc((unsigned)(((keysize_query+datasize_text)/DIV_HACH+1) * sizeof(int)));
  ASSERT(last_tuple_pos_with_dist_hach,"Assemble");
  memset(last_tuple_pos_with_dist_hach,0xff,(unsigned)(((keysize_query+datasize_text)/DIV_HACH+1) * sizeof(int)));
#endif 

#ifdef STATS
  last_clock = clock();
#endif
  

  /* for each chunk element */
  for( i_chunk = 0 ; i_chunk < nbchunks_text ; i_chunk++ ){
    
    /*  
     * [3] clean used tables  and
     *     and set parameters  
     */
    RESET_TAB(last_tuple_pos_with_dist,(unsigned)( MEMOPT_TABSIZE * sizeof(int)),0xff);
    RESET_TAB(last_tuple_ptr_with_dist,(unsigned)( MEMOPT_TABSIZE * sizeof(tuple *)),0x00);
    chunk_keysize_text  =  chunksize_text[i_chunk] - gp_seeds_span_min + 1;
    chunk_data_text    +=  ((i_chunk == 0) ? (0) : (chunksize_text[i_chunk-1]));
    i_base             +=  ((i_chunk == 0) ? (0) : (chunksize_text[i_chunk-1])); 
    
    /* [4] algorithme de creation des tuples chains */

    /* for each text kword  */
    for(feature->i_current = 0 ; feature->i_current < chunk_keysize_text ; feature->i_current ++ )
      for(seed=0;seed<gp_nb_seeds;seed++) {
     
      code      = Key(chunk_data_text,feature->i_current,seed);    /* we take the code of the kmot 'w' located at feature->i_current position on the "Text" */
      index=first_query[seed][code];     /* search the first occurrence of 'w' on the sequence "Query" */
      index_fin=first_query[seed][code+1];
      i_previous=keylist_query[seed][index];
      
      /* for each occurrence of 'w' on the "Query" sequence */
      while(index<index_fin){ 
	
	distance =  feature->i_current + keysize_query - i_previous;
	/* consider all the interesting distances (diagonals) by adding indel bias  0 +1 -1 +2... until we reach a new tuple or the upper indel bound */
	observed_distance = distance;
	delta_distance = 0;
	
#ifndef NO_CACHE
	/* Jump if no match on the breigthour */
	j_hack=last_tuple_pos_with_dist_hach[H_DIST(observed_distance )]; 
	code=((j_hack< feature->i_current - gp_rho_stat)||(j_hack > feature->i_current));
	if (observed_distance%DIV_HACH>borne_max_hash) {
	  j_hack=last_tuple_pos_with_dist_hach[H_DIST(observed_distance)+1]; 
	  code=code&&((j_hack< feature->i_current - gp_rho_stat)||(j_hack > feature->i_current)); 
	}
	if (observed_distance%DIV_HACH<borne_min_hash) {
	  j_hack=last_tuple_pos_with_dist_hach[H_DIST(observed_distance)-1]; 
	  code=code&&((j_hack< feature->i_current - gp_rho_stat)||(j_hack > feature->i_current));
	}	
	if (code==0)
#endif
	while( delta_distance <= 2 * gp_delta_stat + 1 ){
	  
	  /* search for interesting tuples */
	  j =  GET_TAB_MIN(last_tuple_pos_with_dist,observed_distance,gv_i_current - gp_rho_stat);

	  /* if tuples found are respecting criteria (neighboors on diagonal and on sequences) */
	  if (j > feature->i_current -  gp_rho_stat                    && 
	      j <= feature->i_current                                  &&
	      j - observed_distance <= feature->i_current - distance   && 
	      j != -1 ) { 
	    
	    
	    
	    /*---------------------------------------------------------------------*/
	    
	    /* (1) possibly create one instance of tuple [j] 
	     * if this one was not in any list. 
	     */
	    if (EXIST_TAB(last_tuple_ptr_with_dist,observed_distance)<=0){
#ifdef STATS
	      feature->nb_chains_built+=01;
#endif
	      /* create at tuple list element */
	      feature->last_tl->next = CreateTupleList();
	      feature->last_tl = feature->last_tl->next;
	      
	      /* create a tuple [j] and update lists */
	      t = CreateTuple(j,observed_distance,seed);
	      feature->last_tl->first_tuple = t;
	      PUT_TAB(last_tuple_ptr_with_dist,observed_distance,t); 
	    }
	    else
	      
	    
	    /* (2) we search for the last tuple [j] instanciation 
	     */
	      t = (tuple *)GET_TAB(last_tuple_ptr_with_dist,observed_distance) ;
	    
	    
	    /*
	     * (3) iterate tuples and add this new one to the most adapted one
	     */
	    
	    do{
	      
	      if (  distance == t->distance  && feature->i_current - t->occurrence <= gp_seeds_span[seed] ){
		
		/* (3a) possible fusion with this tuple (no instanciation)
		 *	
		 * [inter-tuple overlap]	
		 */ 

		t->size              +=  feature->i_current -  t->occurrence;
		t->lastSeed           = seed;
		if ((t->lastSeed==-1) ||(gp_seeds_span[t->lastSeed]<gp_seeds_span[seed]))
		  t->lastSeed           = seed;
		t->bitweight         += gp_automata[(~(t->overlappedhits<<(feature->i_current -  t->occurrence)))&gp_seeds_bitweight[t->lastSeed]];
		t->overlappedhits    = t->overlappedhits<<(feature->i_current -  t->occurrence) | gp_seeds_bitweight[t->lastSeed];
		t->overlappedhits    &= AUTOMATAMASK;
		t->occurrence         = feature->i_current;		

		
		goto no_maj_last_tuple_ptr_with_dist;
		
	      }
	      
	      
	      /*	parcours de la liste */
	      if(!(t->next))
		break;
	      t = t->next;
	      
	    }while(1);
	    
	    /*
	     * (3d) any fusion has not been possible
	     *
	     * we instanciate the tuple and add this one at the end of the chain.
	     *
	     */
	    
	    t_ = CreateTuple(feature->i_current, distance,seed);
	    t->next = t_;
	    PUT_TAB    (last_tuple_ptr_with_dist,distance,t_);
	    
	    /*
	     * (4) dont reset last_tuple_ptr_with_dist[distance] to NULL
	     * because it has been modified in (3)
	     */
	    goto no_maj_last_tuple_ptr_with_dist;
	    
	  }
	  
	  /*---------------------------------------------------------------------*/
	  
	  
	  /* diagonal walk d,d+1,d-1,... */
	  delta_distance ++;
	  observed_distance = distance + deltashift[delta_distance];
	  observed_distance = MAX(MIN(observed_distance,feature->i_current+keysize_query),feature->i_current+1);
	  /* border effects avoided*/
	  
	}/*  while : diagonal walk */
	
	
	/* reset last tuple instancied to NULL (because not instance is created)
	 */ 
	
	DEL_TAB    (last_tuple_ptr_with_dist,distance);	
		
	
      no_maj_last_tuple_ptr_with_dist:
	
	/* 
	 * update last tuple position  with diagonal = 'distance' 
	 */
#ifdef STATS
	feature->nb_seeds++;
#endif
	PUT_TAB(last_tuple_pos_with_dist,distance,feature->i_current);

#ifndef NO_CACHE
	last_tuple_pos_with_dist_hach[H_DIST(distance)]=feature->i_current;
#endif
	/* iterate 'w' code list before feature->i_current position */
	index++;
        i_previous=keylist_query[seed][index];
      }  
      
      if (!(feature->i_current % ALIGN_EVERY_NB_ITER)){
#ifdef TRACE
	Display_Progress((unsigned)feature->i_current+i_base,(unsigned)(datasize_text-1),feature);    
#endif
#ifdef STATS
	feature->clock_chain += clock() - last_clock;
	last_clock=  clock();
#endif
#ifdef DEBUGASSEMBLE
	DisplayListTupleList(tuple_lists);
#endif	
	AlignAndFree(
		     data_query,        datasize_query,
		     chunk_data_text,   chunksize_text[i_chunk], 
		     0,i_chunk,
		     feature);
#ifdef STATS
	feature->clock_align += clock() - last_clock;
	last_clock   = clock();
#endif 
	
      }
    }/*end of  [3] (keysize_text loop)  */
#ifdef TRACE
    Display_Progress((unsigned)feature->i_current+i_base,(unsigned)(datasize_text),feature);   
#endif
#ifdef STATS
    feature->clock_chain += clock() - last_clock;
    last_clock=  clock();
#endif	
#ifdef DEBUGASSEMBLE
    DisplayListTupleList(tuple_lists);
#endif	


    AlignAndFree(
		 data_query     ,  datasize_query,
		 chunk_data_text,  chunksize_text[i_chunk], 
		 1,i_chunk,
		 feature);

#ifdef STATS		
    feature->clock_align += clock() - last_clock;
    last_clock =  clock();
#endif   	

  }/*end of  [4] (chunk loop) */

  FREE_TAB(last_tuple_pos_with_dist);
  FREE_TAB(last_tuple_ptr_with_dist);
   
  for(seed=0;seed<gp_nb_seeds;seed++) {
    free(keylist_query[seed]);
    free(first_query[seed]);
  }   
  return 0;
}
  

