/*
 *  YASS 1.09
 *  Copyright (C) 2004,2005
 *  YASS Team
 *  Laurent Noe, Gregory Kucherov, Mikhail Roytberg, Christophe Valmir, Steven Corroy
 *  yass@loria.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software contains code derived from the GNU libavl library.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "util.h"
#include "proba.h"



/*****************************
 *
 * Waiting Time Distribution
 *
 *****************************/


/*
 * bound of waiting time
 *  
 */


#ifdef INLINE
inline
#endif
int statistical_bound_of_waiting_time1(double p, int k, double alpha)
{
    double a1 = dpow(p, k);
    double a2 = (1 - p) * a1;
    int x = 0;
    double Sum = 0.00;

    double *last_k_prob =
	(double *) malloc((unsigned) (k + 1) * sizeof(double));
    ASSERT(last_k_prob, waiting_time_distrib_1);


    while (Sum < (1 - alpha)) {

	if (x < k) {
	    last_k_prob[x % (k + 1)] = 0.00;
	} else if (x == k) {
	    last_k_prob[x % (k + 1)] = a1;
	} else if (x <= 2 * k) {
	    last_k_prob[x % (k + 1)] = a2;
	} else {
	    last_k_prob[x % (k + 1)] =
		last_k_prob[(x + k) % (k + 1)] -
		a2 * last_k_prob[x % (k + 1)];
	}
	Sum += last_k_prob[x % (k + 1)];
	x++;
    }
    free(last_k_prob);

    return x;
}


#ifdef INLINE
inline
#endif
int statistical_bound_of_waiting_time2(double p, int k, double alpha)
{
    double p_k = dpow(p, k);
    double qp_k = (1 - p) * p_k;
    int x = 0;
    double Sum_0_xk1 = 0.00;
    double Sum = 0;

    double *last_k_prob =
	(double *) malloc((unsigned) (k + 1) * sizeof(double));
    ASSERT(last_k_prob, waiting_time_distrib_2);

    while (Sum < (1 - alpha)) {
	if (x < k) {
	    last_k_prob[x % (k + 1)] = 0.00;
	} else if (x == k) {
	    last_k_prob[x % (k + 1)] = p_k;
	} else {
	    Sum_0_xk1 += last_k_prob[x % (k + 1)];
	    last_k_prob[x % (k + 1)] = qp_k * (1 - Sum_0_xk1);
	}

	Sum += last_k_prob[x % (k + 1)];
	x++;
    }
    free(last_k_prob);

    return x;
}




/*****************************
 *
 * Random Walk distribution:
 *
 *****************************/


/*
 *
 * table of probabilities to reach 
 * position "i" at step L  
 * with the rdwalk  
 * -> pr pI for left shift
 * -> pr pI for right shift
 * (pr 1-2pI dont move) 
 */


#ifdef INLINE
inline
#endif
double *randomwalk_probability_of_pos1(double pI, int L)
{

    double *C_N_K;		/* Cnk table */
    double *P_for_score;	/* random walk scores */
    double Sum_of_C_N_K = 1;
    double P_dep;

    int i, nb_dep;

    /* (1) allocation and initialisation */

    C_N_K = (double *) malloc((unsigned) (2 * L + 3) * sizeof(double));
    ASSERT(C_N_K, randomwalk_probability_of_pos1);
    for (i = 0; i < 2 * L + 3; i++)
	C_N_K[i] = 0;


    P_for_score =
	(double *) malloc((unsigned) (2 * L + 1) * sizeof(double));
    ASSERT(P_for_score, randomwalk_probability_of_pos1);
    for (i = 0; i < 2 * L + 1; i++)
	P_for_score[i] = 0;



    /* (2) poly computed
     */

    C_N_K[L] = 1;


    for (nb_dep = 0; nb_dep <= L; nb_dep++) {

	P_dep =
	    C(L, nb_dep) * dpow(pI, nb_dep) * dpow((1 - pI), (L - nb_dep));


    /***************************************/
	if (P_dep > 1 || 1) {	/* infinity problems ??? */
	    goto end;
	}


	/* (2.1) proba computed here */
	for (i = -nb_dep; i <= nb_dep; i++) {
	    P_for_score[i + L] += ((C_N_K[i + L]) * P_dep / Sum_of_C_N_K);
	}

	/* (2.2) C_N_K update */
	if (nb_dep < L) {
	    for (i = -nb_dep - 1; i <= nb_dep + 1; i += 2) {
		C_N_K[i + L] = C_N_K[i - 1 + L] + C_N_K[i + 1 + L];
	    }
	    for (i = -nb_dep; i <= nb_dep; i += 2) {
		C_N_K[i + L] = 0;
	    }
	    Sum_of_C_N_K *= 2;
	}
    }

  end:
    /* (3) free unused stuff */
    free(C_N_K);

    return P_for_score;
}


#ifdef INLINE
inline
#endif
double *randomwalk_probability_of_pos2(double pI, int L)
{

    int i, j;
    double a = pI / 2;
    double b = 1 - pI;
    double *poly =
	(double *) malloc((unsigned) (2 * L + 1) * sizeof(double));
    double *polytmp =
	(double *) malloc((unsigned) (2 * L + 1) * sizeof(double));

    ASSERT(poly, randomwalk_probability_of_pos2);
    ASSERT(polytmp, randomwalk_probability_of_pos2);


    /* (1)  table inits */

    for (i = 0; i < 2 * L + 1; i++) {
	poly[i] = 0;
    }

    /* (2) polynom computed
     * [ p*X^2 + b*X + p ] ^ L
     *
     */

    poly[0] = 1;
    for (i = 0; i < L; i++) {

	/* clear tmp buffer */
	for (j = 0; j < (2 * L + 1); j++) {
	    polytmp[j] = 0;
	}

	/* badly have [ p*X^2 + b*X + p ] ^ i+1 from [ p*X^2 + b*X + p ] ^ i */
	for (j = 0; j < (2 * L + 1); j++)
	    polytmp[j] += poly[j] * a;
	for (j = 0; j < (2 * L); j++)
	    polytmp[j + 1] += poly[j] * b;
	for (j = 0; j < (2 * L - 1); j++)
	    polytmp[j + 2] += poly[j] * a;

	/* copy [ p*X^2 + b*X + p ] ^ i+1 into poly[] */
	for (j = 0; j < (2 * L + 1); j++)
	    poly[j] = polytmp[j];

    }				/* i */

    free(polytmp);

    return poly;

}


#ifdef INLINE
inline
#endif
double *randomwalk_probability_of_pos3(double pI, int L)
{

    int i, j;
    double a = pI * 0.50;
    double b = 1 - pI;
    double *u = (double *) malloc((unsigned) (2 * L + 1) * sizeof(double));
    double *f = (double *) malloc((unsigned) (2 * L + 1) * sizeof(double));
    double *t = (double *) malloc((unsigned) (2 * L + 1) * sizeof(double));
    double *s = NULL;
    int P = L;


    ASSERT(u, randomwalk_probability_of_pos3);
    ASSERT(f, randomwalk_probability_of_pos3);
    ASSERT(t, randomwalk_probability_of_pos3);

    /* (1) tables inits */

    for (i = 0; i < 2 * L + 1; i++)
	u[i] = 0;
    for (i = 0; i < 2 * L + 1; i++)
	f[i] = 0;

    f[0] = 1.00;
    u[0] = a;
    u[1] = b;
    u[2] = a;

    /* (2) qpow */
    while (P > 0) {
	if (P & 1) {
	    /* f = f*u */
	    for (i = 0; i < 2 * L + 1; i++) {
		t[i] = 0;
		for (j = 0; j <= i; j++)
		    t[i] += u[i - j] * f[j];
	    }
	    s = t;
	    t = f;
	    f = s;
	}
	/* u = u * u; */
	for (i = 0; i < 2 * L + 1; i++) {
	    t[i] = 0;
	    for (j = 0; j <= i; j++)
		t[i] += u[i - j] * u[j];
	}
	s = t;
	t = u;
	u = s;
	P >>= 1;
    }

    free(u);
    free(t);
    return f;

}







/*
 *
 * give rd walk bounds in  (1-alpha)% 
 * 
 */



#ifdef INLINE
inline
#endif
int statistical_bound_of_randomwalk1(double pI, int L, double alpha)
{
    double *RDW_Bound = randomwalk_probability_of_pos1(pI, L);
    double Sum = RDW_Bound[L];
    int bound = 0;
    while (Sum < (1 - alpha) && bound < L) {
	bound++;
	Sum += RDW_Bound[L - bound] + RDW_Bound[L + bound];
    }

    /* overflow in table */
    if (bound == L)
	bound = (int) (pow((double) pI * L, (double) (1 / 2.3))) + 1;	/* asympt value */

    free(RDW_Bound);
    return bound;
}




#ifdef INLINE
inline
#endif
int statistical_bound_of_randomwalk2(double pI, int L, double alpha)
{
    double *RDW_Bound = randomwalk_probability_of_pos3(pI, L);
    double Sum = RDW_Bound[L];
    int bound = 0;
    while (Sum < (1 - alpha) && bound < L) {
	bound++;
	Sum += RDW_Bound[L - bound] + RDW_Bound[L + bound];
    }

    free(RDW_Bound);
    return bound;
}



/*
 * Scores,Evalue,and Bitscore ...
 *
 */

double
Evalue(double K, double Lambda, int query_size, int text_size, int score)
{
    return K * query_size * text_size * exp(-Lambda * score);
}

double BitScore(double K, double Lambda, int score)
{
    return (Lambda * ((double) score) - log(K)) / log(2);
}

int
MinScore(double K, double Lambda, int query_size, int text_size,
	 double MaxEvalue)
{
    return (int) floor((1 / Lambda) *
		       (log(K * query_size * text_size) - log(MaxEvalue)));
}



/*
 *  trinomial prob
 */

#ifdef INLINE
inline
#endif
double P_mutation_bias(int freq[])
{
    int i;
    int n = freq[0] + freq[1] + freq[2];
    double result = 1.00;

    if (n == 0)
	return 1.00;

    for (i = 1; i <= n; i++) {
	result *= ((double) i) / 3;
    }
    for (i = 1; i <= freq[0]; i++)
	result /= (double) i;
    for (i = 1; i <= freq[1]; i++)
	result /= (double) i;
    for (i = 1; i <= freq[2]; i++)
	result /= (double) i;

    return result;
}

/*
 *
 *  Read a paper Mothods for assessing the statistical signifiance of molecular
 *  sequence features by using general scoring schemes, writting by samuel Karlin
 *  and F. Atschul in 1989-90.
 *
 */



/*
 *
 *  Compute probabilty
 *
 */

void
computeProbability(double *p_probabilite[] /*out tableau de NBRE_SCORE */ ,
		   int *p_score[]          /*out tableau de NBRE_SCORE */ ,
		   int *p_nbreScore /*out */ )
{
    unsigned i;
    double *probabilite;
    int *score;
    int nbreScore;
    /* Nombre de lettre dans chaque sequence */
    /* A changer quand il y a d'autre lettre */
    unsigned int nbL[2] =
	{ gp_nb_letter[0][0] + gp_nb_letter[0][1] + gp_nb_letter[0][2] +
gp_nb_letter[0][3],
	gp_nb_letter[1][0] + gp_nb_letter[1][1] + gp_nb_letter[1][2] +
	    gp_nb_letter[1][3]
    };


    /* 1.1) intialise le nombre de Score */
    nbreScore = (int) NBRE_SCORE;

    /* 1.2) Calcule pour chaque score sa probabilite */
    score = malloc(NBRE_SCORE * sizeof(int));
    probabilite = malloc(NBRE_SCORE * sizeof(double));
    ASSERT(score, calculeProbabilite);
    ASSERT(probabilite, calculeProbabilite);
    score[0] = gp_cost_match;
    score[1] = gp_cost_mismatch;
    score[2] = gp_cost_transition;
    score[3] = gp_cost_other;

    /* 2.1) calcule la probabilite d'un match */
    probabilite[0] = 0;
    for (i = 0; i < 4; i++) {
	probabilite[0] += ((double) gp_nb_letter[0][i] / (double) nbL[0]) *
	    ((double) gp_nb_letter[1][i] / (double) nbL[1]);
    }

    /* 2.2) calcule la probabilite d'une Transition A->G, T->C, G->A, C->T */
    probabilite[1] = 0;
    for (i = 0; i < 2; i++) {
	probabilite[1] += ((double) gp_nb_letter[0][i] / (double) nbL[0]) *
	    ((double) gp_nb_letter[1][i + 2] / (double) nbL[1]);
	probabilite[1] +=
	    ((double) gp_nb_letter[0][i + 2] / (double) nbL[0]) *
	    ((double) gp_nb_letter[1][i] / (double) nbL[1]);
    }
    /* 2.3) calcule la probabilite d'un Transvertion A->C,A->T, T->G,T->A,
       G->C,G->T, C->G,C->A, */
    probabilite[2] = 1.00 - probabilite[0] - probabilite[1];


    /* 2.4) calcule la probabilite d'un Other */
    probabilite[3] = 0.0;

    /* 2.4) Display probability */
    /*   for (i=0;i<4;i++) { 
       fprintf(stderr,"probability [%d]=%g\n",i,probabilite[i]); 
       } */

    /* 3) return parametre */
    *p_probabilite = probabilite;
    *p_score = score;
    *p_nbreScore = nbreScore;
}



#define COMPUTELAMBDA(lambda){                   \
  S=0.0;                                         \
  for (i=0;i<nbrScore;i++){                      \
     S+=proba[i]*exp(lambda*(double)score[i]);   \
  }                                              \
}

/*
 *
 * Compute Lambda
 *
 */
double computeLambda(double *proba,	/* Probabilite match/transition/transvertion/Other */
		     int *score, int nbrScore)
{
    int i;
    double lambda_lower = 0.0;
    double lambda_upper = 1.0;
    double lambda = 0;
    double S = 0;

    /*1) check feasability  */
    for (i = 0; i < 4; i++)
	S += proba[i] * (double) score[i];
    if (S > 0.0) {
	fprintf(stderr,"Lambda value cannot be computed (Please fix it with the -L option)\n");
	exit(1);
    }

    S = 0.0;
    /*2) looking for lambda upper */
    COMPUTELAMBDA(lambda_upper);
    while (S < 1) {
	lambda_lower = lambda_upper;
	lambda_upper *= 2.0;
	COMPUTELAMBDA(lambda_upper);
    }


    /*3) looking for lambda exactly */
    while (lambda_upper - lambda_lower > SENSIBILITY_LAMBDA) {
	lambda = lambda_lower + (lambda_upper - lambda_lower) / 2.0;
	COMPUTELAMBDA(lambda);
	if (S == 1.0)
	    return lambda;
	if (S > 1.0)
	    lambda_upper = lambda;
	else {
	    lambda_lower = lambda;
	}
    }
    return lambda;
}


/*
 *
 *   compute K
 *
 */
double computeK(double *proba,	/* Probabilite match/transition/transvertion/Other */
		int *score, int nbrScore,	/* length  of the tables proba and score */
		double lambda)
{

    double *pr, *pr_new, *pr_tmp;
    double denominator = 0.0;
    double numerateur = 0.0;
    double Pb, Ek, C;
    int x, i, k, Sk, scoreMin = 0, scoreMax = 0;
    int lengthProb, zeroPosition;


    /*
       Tests  values from BLAST:
     */

    /*
       Valeurs de blast Verifies
       score[0]=1;
       score[1]=0;
       score[2]=-3;
       lambda=1.371;


       score[0]=3;
       score[1]=0;
       score[2]=-2;
       lambda=0.271;


       score[0]=3;
       score[1]=0;
       score[2]=-3;
       lambda=0.317;

       p_proba[0]=0.25;
       p_proba[1]=0;
       p_proba[2]=0.75;
     */


    /* Find minScore,MaxScore to compute the tab length */
    for (i = 0; i < nbrScore; i++) {
	if (scoreMin > score[i])
	    scoreMin = score[i];
	if (scoreMax < score[i])
	    scoreMax = score[i];
    }

    lengthProb = (scoreMax - scoreMin +1) * NBLOOPS_K + 1;
    zeroPosition = -scoreMin * NBLOOPS_K;

    /*1) initialisation des tableaux de proba */
    if (lengthProb > 1024*1024*4) {
      fprintf(stderr,"Yass:compute K is impossible because the numbers in -C option est too big\n");
      exit(1);
    }
    
    
    pr = malloc(lengthProb * sizeof(double));
    pr_new = malloc(lengthProb * sizeof(double));
    ASSERT(pr, computeK);
    ASSERT(pr_new, computeK);

    for (i = 0; i < lengthProb; i++) {
	pr[i] = 0.0;
	pr_new[i] = 0.0;
    }

    /* Initial step for K = 0 */
    pr[zeroPosition] = 1.0;
    

    /* 
     * 2) Compute denominator lambda*E[ S_1 * e^(lambda*S_1)]
     */
    for (i = 0; i < nbrScore; i++)
	denominator += proba[i] * score[i] * exp(lambda * score[i]);
    denominator *= lambda;


    /* 
     * 3) numerateur 1/k*SUM (E[e^lambda*S_k;S_k<0]+Prob(S_k>=0),k,1,SENSIBILITY_K)
     */
    for (k = 1; k < NBLOOPS_K; k++) {

	/* 
	 * 3.1) compute a new probability
	 */
	for (x = 0; x < lengthProb; x++)
	  for (i = 0; i < nbrScore; i++) {
	    pr_new[x + score[i]] += pr[x] * proba[i];
	  }


	/* 
	 * 3.2) permutation pr and pr_new
	 */
	pr_tmp = pr;
	pr = pr_new;
	pr_new = pr_tmp;
	for (i = 0; i < lengthProb; i++)
	    pr_new[i] = 0.0;


	Pb = 0.0;		/* Probabilit Prob(S_k) */
	Ek = 0.0;		/* Esperance   E[e^(lambda*S_k);S_k<0 */


	/*
	 * 3.4) esperance on the negatif value (<0)
	 */
	for (i = 0; i < zeroPosition; i++) {
	    Sk = i - zeroPosition;

	    if (exp(lambda * Sk) > DBL_MAX) {
		fprintf(stderr,"Yass: compute K exp too big\n");
		exit(1);
	    }
	    Ek += pr[i] * exp(lambda * (double) Sk);
	}

	/*
	 * 3.5) probabilit on postif value (>=0)
	 */
	for (; i < lengthProb; i++)
	    Pb += pr[i];

	numerateur += (Pb + Ek) / (double) k;
    }

    C = exp(-2 * numerateur) / denominator;
    free(pr);
    free(pr_new);

    return C * (lambda * DELTA_K) / (1 - exp(-lambda * DELTA_K));
}
