/*
 *  YASS 1.09
 *  Copyright (C) 2004,2005
 *  YASS Team
 *  Laurent Noe, Gregory Kucherov, Mikhail Roytberg, Christophe Valmir, Steven Corroy
 *  yass@loria.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software contains code derived from the GNU libavl library.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
/*
 *    Include windows or linux for a thread
 *
 */
#ifdef WIN32
#include <windows.h>
#else
#include <pthread.h>
#endif

/*********************
 *
 * Thread definition
 *
 ***********************/
Feature_assemble feature_assemble[2];

/*
 *  Create Thread
 */
#ifdef  BIPROC
#ifdef WIN32
#define CREATE_THREAD(thread,function,parameters)                                    \
if ((thread=CreateThread(NULL,0,function, (PVOID) parameters,0, (DWORD *)&x))==NULL){\
      fprintf (stderr,"ERROR:YASS Can not create thread\n");                         \
      fprintf (stderr,"Use Yass mono thread");                                       \
      exit(1);                                                                       \
   } 

#else
#define CREATE_THREAD(thread,function,parameters)                   \
  if (pthread_create(&thread,NULL,function,(void *)parameters)) {   \
      fprintf (stderr,"ERROR:YASS Can not create thread\n");        \
      fprintf (stderr,"Use Yass mono thread");                      \
      exit(1);                                                      \
   }
#endif
#else
#define CREATE_THREAD(thread,function,parameters) \
   function(parameters);                                                   
#endif                                                              

/*
 *  End Thread
 */
#ifdef BIPROC
#ifdef WIN32
#define END_THREAD() return 0;
#else
#define END_THREAD() pthread_exit(NULL);
#endif
#else
#define END_THREAD() return NULL;
#endif

/*
 *  Wait Thread
 */
#ifdef BIPROC
#ifdef WIN32
#define WAIT_THREAD(THREAD)                   \
    WaitForSingleObject(THREAD, INFINITE); 
#else
#define WAIT_THREAD(THREAD)                   \
    pthread_join (THREAD, (void *) &x);
#endif
#else
#define WAIT_THREAD(THREAD) ;
#endif

/*
 *  Affichage des grouping
 */
int             regroup_begin  =0;
int             regroup_end    =0;
int             filtering_begin=0;
int             filtering_end  =0;


#ifdef BIPROC
#ifdef WIN32
/* WINDOWS multi proc*/
HANDLE regroup_mutex_begin;
HANDLE regroup_mutex_end  ;
HANDLE filtering_mutex_begin;
HANDLE filtering_mutex_end;


#define INIT_MUTEX()                                                   \
 regroup_mutex_begin=CreateMutex(NULL,FALSE,"regroup_mutex_begin");    \
 regroup_mutex_end  =CreateMutex(NULL,FALSE,"regroup_mutex_end");      \
 filtering_mutex_begin=CreateMutex(NULL,FALSE,"filtering_mutex_begin");\
 filtering_mutex_end=CreateMutex(NULL,FALSE,"filtering_mutex_end");    \


#define DISPLAY_BEGIN(NAMEMUTEX,NOMVAR,STRING)              \
  WaitForSingleObject(NAMEMUTEX,INFINITE);                  \
  NOMVAR++;                                                 \
  if(NOMVAR>=gp_reverse) fprintf(stderr,"%s",STRING);       \
  ReleaseMutex(NAMEMUTEX);                                  \

#define DISPLAY_END(NAMEMUTEX,NOMVAR)                  \
  WaitForSingleObject(NAMEMUTEX,INFINITE);             \
  NOMVAR++;                                            \
  if(NOMVAR>=gp_reverse) fprintf(stderr,"finished\n"); \
  ReleaseMutex(NAMEMUTEX);                             \

#else
/* UNIX multi proc*/
pthread_mutex_t regroup_mutex_begin;
pthread_mutex_t regroup_mutex_end  ;
pthread_mutex_t filtering_mutex_begin;
pthread_mutex_t filtering_mutex_end  ;


#define INIT_MUTEX()  \
 pthread_mutex_init(&regroup_mutex_begin,NULL);    \
 pthread_mutex_init(&regroup_mutex_end,NULL);      \
 pthread_mutex_init(&filtering_mutex_begin,NULL);  \
 pthread_mutex_init(&filtering_mutex_end,NULL);    \




#define DISPLAY_BEGIN(NAMEMUTEX,NOMVAR,STRING)           \
  pthread_mutex_lock(&NAMEMUTEX);                        \
  NOMVAR++;                                              \
  if(NOMVAR>=gp_reverse) fprintf(stderr,"%s",STRING);    \
  pthread_mutex_unlock(&NAMEMUTEX);                      \

#define DISPLAY_END(NAMEMUTEX,NOMVAR)                  \
  pthread_mutex_lock(&NAMEMUTEX);                      \
  NOMVAR++;                                            \
  if(NOMVAR>=gp_reverse) fprintf(stderr,"finished\n"); \
  pthread_mutex_unlock(&NAMEMUTEX);                    \

#endif 
#else
/* Mono Proc*/
int display_regroup_mutex_begin;
int regroup_mutex_begin;
int regroup_mutex_end  ;
int filtering_mutex_begin;
int filtering_mutex_end  ;


#define INIT_MUTEX()  ;

#define DISPLAY_BEGIN(NAMEMUT,NOMVAR,STRING)             \
  NOMVAR++;                                              \
  if(NOMVAR==2) fprintf(stderr,"%s",STRING);             \

#define DISPLAY_END(NAMEMUT,NOMVAR)                      \
  NOMVAR++;                                              \
  if(NOMVAR==2) fprintf(stderr,"finished\n");            \

#endif






